  /********************************************************************
   *                                                                  *
   *  Project: 16F1828 Thermal Switch v1                              *
   *   Source: 16F1828_Thermal_Switch_v1.c                            *
   *   Author: Mike McLaren, K8LH                                     *
   *  (C)2012: Micro Application Consultants                          *
   *     Date: 06-May-2012  Revised: 17-May-2012, 0330UTC             *
   *                                                                  *
   *   16F1828 "Thermal Switch v1" project experiment to test low     *
   *   level Dallas One-Wire drivers for DS1820 series temperature    *
   *   sensors on the K8LH 4-digit 7-segment "demo-1" board.          *
   *                                                                  *
   *      IDE: MPLAB 8.84 (tabs = 4)                                  *
   *     Lang: SourceBoost BoostC v7.05, Lite/Free version            *
   *                                                                  *
   ********************************************************************/

   #include <system.h>

   #pragma DATA _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _MCLRE_OFF
   #pragma DATA _CONFIG2, _PLLEN_OFF & _LVP_OFF

   #pragma CLOCK_FREQ 8000000       // 8 MHz INTOSC
   #pragma OPTIMIZE "a"

  /********************************************************************
   *  function prototypes                                             *
   ********************************************************************/

  /********************************************************************
   *  typedef and defines                                             *
   ********************************************************************/

      typedef unsigned char u08;
      typedef unsigned int u16;
      typedef signed short s16;

      #define r08 const rom unsigned char
   
      #define piezo lata.1         // piezo speaker (RA1)

  /*
   *  defines for One-Wire interface on RA0 pin
   */
      #define owport _porta     // owpin PORTA
      #define owtris _trisa     // owpin TRISA
      #define owpin 0           // owpin index for RA0 pin

  /*
   *  DS18B20 command constants
   */
      #define OwSkipRom 0xCC
      #define OwReadRom 0x33
      #define OwConvert 0x44
      #define OwScratch 0xBE

  /*
   *  assembly language helpers
   */
      #define clrc  bcf   _status,C
      #define setc  bsf   _status,C
      #define skpz  btfss _status,Z
      #define skpnz btfsc _status,Z
      #define skpc  btfss _status,C
      #define skpnc btfsc _status,C

  /********************************************************************
   *  variables (and const arrays)                                    *
   ********************************************************************/

  /*
   *  One-Wire driver and related function variables (10 bytes)
   */
      s16 temp16 @0x20;            // raw 16C temperature word
      u08 templo @0x20;            // raw 16C lo byte
      u08 temphi @0x21;            // raw 16C hi byte
      s16 bin;                     // 16 bit binary work variable
      u08 frac;                    // fraction bits
      u08 family;                  // family ID byte from ROM ID
      u08 thou;                    // bin-to-bcd output
      u08 huns;                    //  "
      u08 tens;                    //  "
      u08 ones;                    //  "

  /*
   *  7-segment display & switch management variables (10 bytes)
   */
      u08 colsel = 0x10;        // display column select mask
      u08 segmask = 0x01;       // display segment select mask
      u08 segdata;              // display segment pattern
      u08 flash;                // display flash timer

      u08 displo = 0x00;        // packed bcd lo digits
      u08 disphi = 0x00;        // packed bcd hi digits

      u08 swnew = 0;            // switch state logic work var
      u08 swold = 0;            // switch state latch
      u08 flags = 0;            // switch flags (bits 7..4)
      u08 beepctr = 0;          // beep task var
   
      #define negflag flags.0   // negative temperature flag
   
      r08 segment[] = {
                 0b00111111,    // "0"   -|-|F|E|D|C|B|A
                 0b00000110,    // "1"   -|-|-|-|-|C|B|-
                 0b01011011,    // "2"   -|G|-|E|D|-|B|A
                 0b01001111,    // "3"   -|G|-|-|D|C|B|A
                 0b01100110,    // "4"   -|G|F|-|-|C|B|-
                 0b01101101,    // "5"   -|G|F|-|D|C|-|A
                 0b01111101,    // "6"   -|G|F|E|D|C|-|A
                 0b00000111,    // "7"   -|-|-|-|-|C|B|A
                 0b01111111,    // "8"   -|G|F|E|D|C|B|A
                 0b01101111,    // "9"   -|G|F|-|D|C|B|A
                 0b01110111,    // "A"   -|G|F|E|-|C|B|A
                 0b01111100,    // "b"   -|G|F|E|D|C|-|-
                 0b00111001,    // "C"   -|-|F|E|D|-|-|A
                 0b01011110,    // "d"   -|G|-|E|D|C|B|-
                 0b01111001,    // "E"   -|G|F|E|D|-|-|A
                 0b01110001,    // "F"   -|G|F|E|-|-|-|A
                 0b00000000,    // " "   -|-|-|-|-|-|-|-
                 0b01000000 };  // "-"   -|G|-|-|-|-|-|-


  /********************************************************************
   *  functions                                                       *
   ********************************************************************/

  /*                                                                  *
   *  (C) 2012, K8LH - One-Wire "Reset" function                      *
   *                                                                  */
   void owreset()               //
   { intcon.GIE = 0;            // interrupts off
     asm bcf    owtris,owpin    // set owpin as output
     asm bcf    owport,owpin    // set owpin to output '0'
     intcon.GIE = 1;            // interrupts on
     delay_100us(5);            //
     asm bsf    owtris,owpin    // set owpin to hi-z '1'
     delay_100us(5);            //
   }                            //

  /*                                                                  *
   *  (C) 2012, K8LH - low-level One-Wire Rd/Wr (Carry) Bit Driver    *
   *  (hard coded for "owport/owtris/owpin" and an 8-MHz clock).      *
   *                                                                  */
   void owrw()                  // read/write Carry bit
   { intcon.GIE = 0;            // interrupts off
   //asm movlb  owport          // bank 0 (inserted by compiler)
     asm bcf    owport,owpin    // clear owpin output latch
   //asm movlb  owtris          // bank 1 (inserted by compiler)
     asm bcf    owtris,owpin    // set owpin to output '0'
     asm movf   _pcl,F          // pseudo bra $+1 (2 cycles)
     asm movf   _pcl,F          // pseudo bra $+1 (2 cycles)
     asm movf   _pcl,F          // pseudo bra $+1 (2 cycles)
     asm skpnc                  // if a '1' (C=1)
     asm bsf    owtris,owpin    // set owpin (RA0) to hi-z '1'
     asm movlw  18/3            //
   dly1:
     asm decfsz _wreg,W         // DelayCy(14*usecs-10)
     asm bra    dly1            //
   //asm movlb  owport          // bank 0 (inserted by compiler)
     asm btfss  owport,owpin    // owpin = 0? no, skip, else
     asm clrc                   // set C = 0
     asm movlw  93/3            // balance of 60-usec "r/w slot"
   dly2:
     asm decfsz _wreg,W         // DelayCy(46*usecs-3)
     asm bra    dly2            //
   //asm movlb  owtris          // bank 1 (inserted by compiler)
     asm bsf    owtris,owpin    // set owpin (RA0) to hi-z '1'
     intcon.GIE = 1;            // interrupts on
   }                            //

  /*                                                                  *
   *  (C) 2012, K8LH - low-level One-Wire Rd/Wr Byte Driver           *
   *                                                                  */
   char owrw(u08 pbyte)         // read/write byte
   { u08 i = 8;                 // bit counter
     do                         //
     { asm rrf  _pbyte,W        // put output bit b0 in Carry
       owrw();                  // read/write bit
       asm rrf  _pbyte,F        // save input bit from Carry
     } while(--i);              //
     return pbyte;              //
   }                            //

  /********************************************************************
   *  main init                                                       *
   ********************************************************************/

   void main()
   {
     ansela = 0;                // ADC off for digital I/O
     anselb = 0;                //  "   "   "    "
     anselc = 0;                //  "   "   "    "
     porta = 0;                 //
     portb = 0;                 //
     portc = 0;                 //
     trisa = 1 << owpin;        // set owpin (RA0) to input
     trisb = 0;                 //
     trisc = 0;                 //
     option_reg.NOT_WPUEN = 0;  // enable weak pull-ups
     wpua = 1 << owpin;         // owpin pull-up enabled

     osccon = 0b01110000;       // set INTOSC to 8 MHz
     while(!oscstat.HFIOFS);    // wait for oscillator stable
  /*                                                                  *
   *  250-us (500 cycle) TMR2 interrupts for a 125-Hz refresh rate    *
   *                                                                  */
     tmr2 = 0;                  // clear Timer 2 register
     t2con = 0b00000001;        // '0-------' unimplemented bit
                                // '-0000---' TOUTPS<3:0>, post 1
                                // '-----0--' TMR2ON, Timer 2 off
                                // '------01' T2CKPS<1:0>, pre 4
     pr2 = 125-1;               // 125 x 2-usec 'ticks' = 250 usecs
     pir1 = 0;                  // clear peripheral interrupt flags
     pie1.TMR2IE = 1;           // set Timer 2 interrupt enable bit
     intcon = 0b11000000;       // '1-------' GIE, enable global and
                                // '-1------' PEIE, peripheral ints
                                // '--0-----' T0IE, TMR0 ints off
                                // '---0----' INTE, off
                                // '----0---' GPIE, IOC disabled
                                // '-----000' T0IF/INTF/GPIF flags
     t2con.TMR2ON = 1;          // start TMR2 (and interrupts)

  /********************************************************************
   *  main loop                                                       *
   ********************************************************************/
     while(1)                   //
     {                          //
       owreset();               // reset one-wire device
       owrw(OwReadRom);         // send "read rom" command
       family = owrw(0xFF);     // get family id byte
       owreset();               // reset one-wire device
       owrw(OwSkipRom);         // skip rom id
       owrw(OwConvert);         // start temperature conversion
       delay_ms(2);             // wait for conversion to complete
       owreset();               // reset one-wire device
       owrw(OwSkipRom);         // skip rom id
       owrw(OwScratch);         // send "read scratchpad" command
       templo = owrw(0xFF);     // read temperature lo byte
       temphi = owrw(0xFF);     // read temperature hi byte

    /*                                                                *
     *  convert DS1820 or DS18S20 9-bit temperature reading (1/2     *
     *  resolution) to DS18B20 12-bit format (1/16th resolution).    *
     *                                                                */
       if(family == 0x10)       // if DS1820 or DS18S20 family ID
       { owrw(0xFF);            //
         owrw(0xFF);            //
         owrw(0xFF);            //
         owrw(0xFF);            //
         frac = 16-owrw(0xFF);  // frac = 16 - 'count_remain', 0..15
         temp16.0 = 0;          // clear the 0.5 bit
         temp16 <<= 3;          // make room for fraction bits
         templo |= frac;        // insert the four fraction bits
         temp16 -= 4;           // subtract 1/4 (per datasheet)
       }

    /*                                                                *
     *  the 16C term in the formulas below is the raw Celsius * 16    *
     *  temperature data from the DS18B20 in the 'templo' variable.   *
     *  a value of C*10 or F*10 is used for the display with a      *
     *  decimal point inserted before the last (tenths) digit.        *
     *                                                                *
     *  10C = 8C + 2C = 16C/2 + (16C+4)/8                             *
     *                                                                *
     *  10F = 18C + 320 = 16C + 2C + 320 = 16C + (16C+4)/8 + 320      *
     *                                                                */
       bin = temp16;            // calculate common '2C' term first
       bin += 4;                // rounding
       bin /= 8;                // now bin = '2C' (with rounding)
       if(flags.7)              // Rt arrow flag = 1 --> Fahrenheit
       { bin += temp16;         // now bin = 16C + 2C
         bin += 320;            // now bin = 16C + 2C + 320 = 10F
       }                        //
       else                     // Rt arrow flag = 0 --> Celsius
       { temp16 /= 2;           // templo = 8C
         bin += temp16;         // now bin = 8C + 2C = 10C
       }                        //
       if(negflag = bin.15)     // set/clear "negflag", if negative
         bin = -bin;            // twos complement the temperature
    /*                                                                *
     *  crude binary-to-decimal routine                               *
     *                                                                */
       asm {                    //
       b2d9999:
         clrf    _thou          //
         movlw   -10            // wreg = -10
       resHuns:
         movwf   _huns          // set huns = -10
       resTens:
         movwf   _tens          // set tens = -10
       div10z:
         addwf   _bin+0,F       // bin = bin - 10, borrow?
         btfss   _status,C      // no, skip, else
         decf    _bin+1,F       // decrement hi byte
         btfsc   _bin+1,7       // negative? no, skip, else
         bra     b2dwrap        // wrap up (borrow/underflow)
         incfsz  _tens,F        // tens overflow? yes, skip, else
         bra     div10z         // loop
         incfsz  _huns,F        // huns overflow? yes, skip, else
         bra     resTens        // reset 'tens' & loop
         incf    _thou,F        // bump 'thou'
         bra     resHuns        // reset 'huns' & 'tens' & loop
       b2dwrap:
         subwf   _huns,F        // fix huns
         subwf   _tens,F        // fix tens
         subwf   _bin+0,W       // fix ones
         movwf   _ones          //
     //
     //  update the display buffers
     //
         swapf  _thou,W         //
         iorwf  _huns,W         //
         movwf  _disphi         // update packed bcd 'disphi'
         swapf  _tens,W         //
         iorwf  _ones,W         //
         movwf  _displo         // update packed bcd 'displo'
       }                        //

       delay_ms(250);           //
       delay_ms(250);           //
       delay_ms(250);           //
     }
   }

  /********************************************************************
   *  interrupt                                                       *
   ********************************************************************/

   void interrupt()             // 250-us (500 cyc) TMR2 interrupts
   { u08 index;                 //
     pir1.TMR2IF = 0;           // clear TMR2 interrupt flag
    /*                                                                *
     *  refresh display (display new segment)                         *
     *                                                                */
     portb = 0;                 // blank the display
     portc = ~segdata;          // new segment (invert, active lo)
     portb = colsel;            // light new segment (column "on")

    /*                                                                *
     *  beep task generates a 32-ms 2000-Hz "new press" beep          *
     *                                                                */
     if(beepctr)                // if beep task running
     { piezo ^= 1;              // toggle piezo spkr pin (RA1) 
       beepctr--;               //
     }                          //

    /*                                                                *
     *  K8LH Parallel Switch State Logic (8 ms debounce intervals)    *
     *                                                                *
     *  swnew  ___---___---______  sample active hi switches          *
     *  swold  ____---___---_____  switch state latch                 *
     *  swnew  ___-__-__-__-_____  changes, press or release          *
     *  swnew  ___-_____-________  filter out 'release' bits          *
     *  flags  ___-------________  toggle flag bits for main          *
     *                                                                */
     if(segmask.0)              // if segment A (2-ms) interval
     { swnew = swold | colsel;  // indicate pressed unconditionally
       if(porta.3 == 0)         // if not pressed (RA3 = 0)
         swnew ^= colsel;       // indicate released
       swnew ^= swold;          // changes, press or release
       swold ^= swnew;          // update switch state latch
       swnew &= swold;          // filter out 'release' bits
       if(swnew)                // if any "new press" bits
         beepctr = 128;         // task a "new press" beep
       flags ^= swnew;          // toggle flag bits for main
     }                          //

    /*                                                                *
     *  advance segment mask bit for next interrupt and if last       *
     *  segment, bump column select bit mask.                         *
     *                                                                */
     segmask <<= 1;             // 
     if(segmask == 0)           // if last segment (2-ms interval)
     { segmask = 1;             // reset to segment A and
       colsel <<= 1;            // advance column select bit
       if(colsel == 0)          // if last column
         colsel = 0x10;         // reset to first column (RB4)
       flash++;                 // bump 'flash' timer
     }                          //

    /*                                                                *
     *  prepare new segment data for next interrupt                   *
     *                                                                */
     index = disphi >> 4;       // 0..9
     if(!index) index = 16;     // suppress leading zero
     if(negflag) index = 17;    // index for "-" character
     if(colsel.5)               // 
     { index = disphi & 15;     //
       if(!disphi) index = 16;  // suppress leading zero
     }                          //
     if(colsel.6) index = displo >> 4;
     if(colsel.7) index = displo & 15;
     segdata = segment[index];  // get segment data
     if(colsel.6)               // if "ones" digit
       segdata.7 = 1;           // decimal point on
     segdata &= segmask;        // only one segment each interrupt

    /*                                                                *
     *  flash entire display at 2-Hz rate in set mode                 *
     *                                                                */
     if(flags.4)                // if SET switch "on" and
     { if(flash.7)              // if 256-ms flash interval
         segdata &= 0x80;       // turn the segment off
     }                          //
   }                            //

